"""Statistics on the bot alert rate."""

import collections
import functools

# INIT ########################################################################

def init_alert_history(size: int) -> collections.deque:
    """Creates a FIFO of fixed size N to store the alerts for the latest N transactions."""
    return collections.deque(size * [()], size)

# UPDATE ######################################################################

def update_alert_history(fifo: collections.deque, alerts: tuple) -> None:
    """Push the alert ids for the latest block into the history."""
    fifo.append(alerts)

# PROCESS #####################################################################

def calculate_alert_rate(fifo: collections.deque, alert: str) -> float:
    """Calculate the alert rate for a given id over the last N transactions."""
    return min(sum([__t.count(alert) for __t in fifo]) / len(fifo), 1.)

# WRAPPER #####################################################################

def alert_history(size: int) -> callable:
    """Creates a decorator for handle_transaction to add an alert history."""
    __history = init_alert_history(size=size)

    def __decorator(func: callable) -> callable:
        """Actually wraps the handle_transaction and handles the alert history."""

        @functools.wraps(func)
        def __wrapper(*args, **kwargs):
            """Main function called on the logs gathered by the Forta network."""
            __findings = func(*args, **kwargs)
            update_alert_history(fifo=__history, alerts=tuple(__f.alert_id for __f in __findings))
            for __f in __findings:
                __f.metadata['anomaly_score'] = calculate_alert_rate(fifo=__history, alert=__f.alert_id)
            return __findings

        return __wrapper

    return __decorator



"""Load the bot metadata."""

import json
import logging
import os

import web3

# BOT VERSION #################################################################

def get_bot_version() -> str:
    """Read the bot version from the package metadata."""
    __version = ''
    try:
        with open('package.json', 'r') as __f:
            __metadata = json.load(__f)
            __version = __metadata.get('version', '')
    except Exception as e:
        logging.error('could not find the file "package.json"')
    finally:
        return __version

# CHAIN ID ####################################################################

def load_chain_id(provider: web3.Web3) -> int:
    """Load the chain id into the execution environment."""
    __chain_id = 1
    try:
        __chain_id = int(os.environ.get('FORTA_CHAIN_ID', '') or provider.eth.chain_id)
        os.environ['FORTA_CHAIN_ID'] = str(__chain_id)
        logging.info(f'set chain id to {__chain_id}')
    except Exception as e:
        logging.error(f'error getting the chain id (kept to {__chain_id})')
    finally:
        return __chain_id

# SECRETS #####################################################################

def load_secrets(filepath: str='secrets.json') -> dict:
    """Load the secrets into the execution environment."""
    __secrets = {}
    try:
        with open(filepath, 'r') as __f:
            __secrets = json.load(__f)
            for __k, __v in __secrets.items():
                os.environ[__k] = __v
    except Exception as e:
        logging.info(f'could not find the file "{filepath}"')
    finally:
        return __secrets



"""Log scanning events."""

import logging
import sys

import forta_toolkit.metadata

# FORMAT ######################################################################

MESSAGE_PATTERN = '[{version}{{levelname}}] {{message}}'

def setup_log_format(pattern: str=MESSAGE_PATTERN, version: str='') -> str:
    """Return the log format string with the common informations filled."""
    __version = version
    # try reading the package metadata
    if not __version:
        __version = forta_toolkit.metadata.get_bot_version()
    # include the bot version in the log message if known
    if __version:
        __version = __version + ' - '
    return pattern.format(version=__version)

# LOGGING #####################################################################

def setup_logger(level: int=logging.INFO, pattern: str=setup_log_format(pattern=MESSAGE_PATTERN, version='')) -> None:
    """Configure the default log objects for a specific bot."""
    __formatter = logging.Formatter(pattern, style='{')

    __handler = logging.StreamHandler(sys.stdout)
    __handler.setLevel(level)
    __handler.setFormatter(__formatter)

    __logger = logging.getLogger()
    __logger.setLevel(level)
    __logger.addHandler(__handler)



"""Format data."""

import eth_utils.address

# GENERIC #####################################################################

def strip_hex_prefix(data: str) -> str:
    return data.replace('0x', '')

# ADDRESS #####################################################################

def format_address_with_checksum(address: str) -> str:
    __address = '0x{0:0>40x}'.format(int(address if address else '0', 16))
    return eth_utils.address.to_checksum_address(__address)
